# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
from datetime import datetime

import pytest
from factory import Factory, Faker, LazyFunction, Sequence, SubFactory, faker
from pytest_factoryboy import register

from correios.models import data
from correios.models.address import Address, ReceiverAddress, SenderAddress
from correios.models.data import TRACKING_PREFIX
from correios.models.posting import (
    Package,
    PostalUnit,
    PostInfo,
    PostingList,
    Receipt,
    ShippingLabel,
    TrackingCode,
    TrackingEvent,
)
from correios.models.user import Contract, FederalTaxNumber, PostingCard, StateTaxNumber, User

try:
    from correios import client as correios
except ImportError:
    correios = None


@pytest.fixture
def valid_federal_tax_number():
    return FederalTaxNumber("73.119.555/0001-20")


@pytest.fixture
def valid_state_tax_number():
    return StateTaxNumber("73.119.555/0001-20")


class UserFactory(Factory):
    class Meta:
        model = User

    name = "ECT"
    federal_tax_number = "34028316000103"
    state_tax_number = "0733382100116"
    status_number = 1


register(UserFactory, "user")


class ContractFactory(Factory):
    class Meta:
        model = Contract

    user = SubFactory(UserFactory)
    number = 9911222777
    regional_direction = 10


register(ContractFactory, "contract")


class PostingCardFactory(Factory):
    class Meta:
        model = PostingCard

    contract = SubFactory(ContractFactory)
    number = 56789123
    administrative_code = 8888888


register(PostingCardFactory, "posting_card")


class TrackingEventFactory(Factory):
    class Meta:
        model = TrackingEvent

    timestamp = datetime(2016, 1, 1, 12)
    status = ('PO', '1')
    location_zip_code = "07192-100"
    location = "CEE"
    city = faker.Faker("city", locale="pt_BR")
    state = faker.Faker("estado_sigla", locale="pt_BR")
    description = "Objeto postado"


register(TrackingEventFactory, "tracking_event")


def _random_tracking_code():
    prefix = random.choice(list(TRACKING_PREFIX.keys()))
    number = "".join(str(random.randrange(0, 10)) for _ in range(8))
    return "{}{} BR".format(prefix, number)


class TrackingCodeFactory(Factory):
    class Meta:
        model = TrackingCode

    code = LazyFunction(_random_tracking_code)


register(TrackingCodeFactory, "tracking_code")


class AddressFactory(Factory):
    class Meta:
        model = Address

    name = Faker("name", locale="pt_BR")
    street = Faker("street_name", locale="pt_BR")
    number = Faker("building_number", locale="pt_BR")
    city = Faker("city", locale="pt_BR")
    state = Faker("estado_sigla", locale="pt_BR")
    zip_code = "07192-100"
    complement = Faker("secondary_address")
    neighborhood = Sequence(lambda n: "Neighborhood #{}".format(n))
    phone = Faker("phone_number", locale="pt_BR")
    cellphone = Faker("phone_number", locale="pt_BR")
    email = Faker("email")
    latitude = Faker("latitude", locale="pt_BR")
    longitude = Faker("longitude", locale="pt_BR")


class ReceiverAddressFactory(AddressFactory):
    class Meta:
        model = ReceiverAddress


class SenderAddressFactory(AddressFactory):
    class Meta:
        model = SenderAddress


register(AddressFactory, "address")
register(ReceiverAddressFactory, "receiver_address")
register(SenderAddressFactory, "sender_address")

_services = [
    data.SERVICE_PAC,
    data.SERVICE_SEDEX,
    data.SERVICE_SEDEX10,
    data.SERVICE_SEDEX12,
]


class PackageFactory(Factory):
    class Meta:
        model = Package

    package_type = Package.TYPE_BOX
    width = LazyFunction(lambda: random.randint(11, 30))
    height = LazyFunction(lambda: random.randint(2, 30))
    length = LazyFunction(lambda: random.randint(18, 30))
    weight = LazyFunction(lambda: random.randint(1, 100) * 100)
    service = LazyFunction(lambda: random.choice(_services))
    sequence = Sequence(lambda n: (n, n + 1))


register(PackageFactory, "package")


class ReceiptFactory(Factory):
    class Meta:
        model = Receipt

    number = faker.Faker('pyint')
    post_date = faker.Faker('date', pattern='%Y%m%d')
    value = faker.Faker(
        'pydecimal',
        left_digits=2,
        right_digits=2,
        positive=True
    )


register(ReceiptFactory, 'receipt')


class ShippingLabelFactory(Factory):
    class Meta:
        model = ShippingLabel

    posting_card = SubFactory(PostingCardFactory)
    sender = LazyFunction(SenderAddressFactory.build)
    receiver = LazyFunction(ReceiverAddressFactory.build)
    service = data.SERVICE_PAC
    tracking_code = SubFactory(TrackingCodeFactory)
    package = SubFactory(PackageFactory)
    invoice_number = LazyFunction(lambda: "{!s:>04}".format(random.randint(1234, 9999)))
    order = LazyFunction(lambda: "OLT123ABC{!s:>03}".format(random.randint(1, 999)))
    text = Faker("text", max_nb_chars=100)
    latitude = 0.0
    longitude = 0.0


register(ShippingLabelFactory, "shipping_label")


class PostedShippingLabelFactory(ShippingLabelFactory):
    receipt = SubFactory(ReceiptFactory)


register(PostedShippingLabelFactory, 'posted_shipping_label')


class PostingListFactory(Factory):
    class Meta:
        model = PostingList

    custom_id = Sequence(lambda n: n)


register(PostingListFactory, 'posting_list')


class PostalUnitFactory(Factory):
    class Meta:
        model = PostalUnit

    code = faker.Faker('random_number', digits=4)
    description = Faker('company', locale='pt_BR')


register(PostalUnitFactory, 'postal_unit')


class PostInfoFactory(Factory):
    class Meta:
        model = PostInfo

    postal_unit = SubFactory(PostalUnitFactory)
    posting_list = SubFactory(PostingListFactory)
    value = faker.Faker(
        'pydecimal',
        left_digits=2,
        right_digits=2,
        positive=True
    )


register(PostInfoFactory, 'post_info')


@pytest.fixture
def client():
    return correios.Correios(username="sigep", password="XXXXXX", environment=correios.Correios.TEST)



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal
from typing import Dict, Tuple  # noqa

from ..utils import RangeSet

TRACKING_PREFIX = {
    "AL": "Agentes de leitura",
    "AR": "Avisos de recebimento",
    "AS": "PAC - Ação Social",
    "CA": "Encomenda Internacional - Colis",
    "CB": "Encomenda Internacional - Colis",
    "CC": "Encomenda Internacional - Colis",
    "CD": "Encomenda Internacional - Colis",
    "CE": "Encomenda Internacional - Colis",
    "CF": "Encomenda Internacional - Colis",
    "CG": "Encomenda Internacional - Colis",
    "CH": "Encomenda Internacional - Colis",
    "CI": "Encomenda Internacional - Colis",
    "CJ": "Encomenda Internacional - Colis",
    "CK": "Encomenda Internacional - Colis",
    "CL": "Encomenda Internacional - Colis",
    "CM": "Encomenda Internacional - Colis",
    "CN": "Encomenda Internacional - Colis",
    "CO": "Encomenda Internacional - Colis",
    "CP": "Encomenda Internacional - Colis",
    "CQ": "Encomenda Internacional - Colis",
    "CR": "Carta registrada sem Valor Declarado",
    "CS": "Encomenda Internacional - Colis",
    "CT": "Encomenda Internacional - Colis",
    "CU": "Encomenda internacional - Colis",
    "CV": "Encomenda Internacional - Colis",
    "CW": "Encomenda Internacional - Colis",
    "CX": "Encomenda internacional - Colis ou Selo Lacre para Caixetas",
    "CY": "Encomenda Internacional - Colis",
    "CZ": "Encomenda Internacional - Colis",
    "DA": "SEDEX ou Remessa Expressa com AR Digital",
    "DB": "SEDEX ou Remessa Expressa com AR Digital (Bradesco)",
    "DC": "Remessa Expressa CRLV/CRV/CNH e Notificações",
    "DD": "Devolução de documentos",
    "DE": "Remessa Expressa Talão/Cartão com AR",
    "DF": "e-SEDEX",
    "DG": "SEDEX",
    "DI": "SEDEX ou Remessa Expressa com AR Digital (Itaú)",
    "DJ": "SEDEX",
    "DK": "PAC Extra Grande",
    "DL": "SEDEX",
    "DM": "e-SEDEX",
    "DN": "SEDEX",
    "DO": "SEDEX ou Remessa Expressa com AR Digital (Itaú)",
    "DP": "SEDEX Pagamento na Entrega",
    "DQ": "SEDEX ou Remessa Expressa com AR Digital (Santander)",
    "DR": "Remessa Expressa com AR Digital (Santander)",
    "DS": "SEDEX ou Remessa Expressa com AR Digital (Santander)",
    "DT": "Remessa econômica com AR Digital (DETRAN)",
    "DU": "e-SEDEX",
    "DV": "SEDEX c/ AR digital",
    "DW": "Encomenda SEDEX (Etiqueta Lógica)",
    "DX": "SEDEX 10",
    "EA": "Encomenda Internacional - EMS",
    "EB": "Encomenda Internacional - EMS",
    "EC": "PAC",
    "ED": "Packet Express",
    "EE": "Encomenda Internacional - EMS",
    "EF": "Encomenda Internacional - EMS",
    "EG": "Encomenda Internacional - EMS",
    "EH": "Encomenda Internacional - EMS ou Encomenda com AR Digital",
    "EI": "Encomenda Internacional - EMS",
    "EJ": "Encomenda Internacional - EMS",
    "EK": "Encomenda Internacional - EMS",
    "EL": "Encomenda Internacional - EMS",
    "EM": "Encomenda Internacional - SEDEX Mundi",  # BR Suffix
    # "EM": "Encomenda Internacional - EMS Importação",
    "EN": "Encomenda Internacional - EMS",
    "EO": "Encomenda Internacional - EMS",
    "EP": "Encomenda Internacional - EMS",
    "EQ": "Encomenda de serviço não expressa (ECT)",
    "ER": "Objeto registrado",
    "ES": "e-SEDEX ou EMS",
    "ET": "Encomenda Internacional - EMS",
    "EU": "Encomenda Internacional - EMS",
    "EV": "Encomenda Internacional - EMS",
    "EW": "Encomenda Internacional - EMS",
    "EX": "Encomenda Internacional - EMS",
    "EY": "Encomenda Internacional - EMS",
    "EZ": "Encomenda Internacional - EMS",
    "FA": "FAC registrado",
    "FE": "Encomenda FNDE",
    "FF": "Objeto registrado (DETRAN)",
    "FH": "FAC registrado com AR Digital",
    "FM": "FAC monitorado",
    "FR": "FAC registrado",
    "IA": "Logística Integrada (agendado / avulso)",
    "IC": "Logística Integrada (a cobrar)",
    "ID": "Logística Integrada (devolução de documento)",
    "IE": "Logística Integrada (Especial)",
    "IF": "CPF",
    "II": "Logística Integrada (ECT)",
    "IK": "Logística Integrada com Coleta Simultânea",
    "IM": "Logística Integrada (Medicamentos)",
    "IN": "Correspondência e EMS recebido do Exterior",
    "IP": "Logística Integrada (Programada)",
    "IR": "Impresso Registrado",
    "IS": "Logística integrada standard (medicamentos)",
    "IT": "Remessa Expressa Medicamentos / Logística Integrada Termolábil",
    "IU": "Logística Integrada (urgente)",
    "IX": "EDEI Expresso",
    "JA": "Remessa econômica com AR Digital",
    "JB": "Remessa econômica com AR Digital",
    "JC": "Remessa econômica com AR Digital",
    "JD": "Remessa econômica Talão/Cartão",
    "JE": "Remessa econômica com AR Digital",
    "JF": "Remessa econômica com AR Digital",
    "JG": "Objeto registrado urgente/prioritário",
    "JH": "Objeto registrado urgente / prioritário",
    "JI": "Remessa econômica Talão/Cartão",
    "JJ": "Objeto registrado (Justiça)",
    "JK": "Remessa econômica Talão/Cartão",
    "JL": "Objeto registrado",
    "JM": "Mala Direta Postal Especial",
    "JN": "Objeto registrado econômico",
    "JO": "Objeto registrado urgente",
    "JP": "Receita Federal",
    "JQ": "Remessa econômica com AR Digital",
    "JR": "Objeto registrado urgente / prioritário",
    "JS": "Objeto registrado",
    "JT": "Objeto Registrado Urgente",
    "JV": "Remessa Econômica (c/ AR DIGITAL)",
    "LA": "SEDEX com Logística Reversa Simultânea em Agência",
    "LB": "e-SEDEX com Logística Reversa Simultânea em Agência",
    "LC": "Objeto Internacional (Prime)",
    "LE": "Logística Reversa Econômica",
    "LF": "Objeto Internacional (Prime)",
    "LI": "Objeto Internacional (Prime)",
    "LJ": "Objeto Internacional (Prime)",
    "LK": "Objeto Internacional (Prime)",
    "LM": "Objeto Internacional (Prime)",
    "LN": "Objeto Internacional (Prime)",
    "LP": "PAC com Logística Reversa Simultânea em Agência",
    "LS": "SEDEX Logística Reversa",
    "LV": "Logística Reversa Expressa",
    "LX": "Packet Standard / Econômica",
    "LZ": "Objeto Internacional (Prime)",
    "MA": "Serviços adicionais do Telegrama",
    "MB": "Telegrama (balcão)",
    "MC": "Telegrama (Fonado)",
    "MD": "SEDEX Mundi (Documento interno)",
    "ME": "Telegrama",
    "MF": "Telegrama (Fonado)",
    "MK": "Telegrama (corporativo)",
    "ML": "Fecha Malas (Rabicho)",
    "MM": "Telegrama (Grandes clientes)",
    "MP": "Telegrama (Pré-pago)",
    "MR": "AR digital",
    "MS": "Encomenda Saúde",
    "MT": "Telegrama (Telemail)",
    "MY": "Telegrama internacional (entrante)",
    "MZ": "Telegrama (Correios Online)",
    "NE": "Tele Sena resgatada",
    "NX": "EDEI Econômico (não urgente)",
    "PA": "Passaporte",
    "PB": "PAC",
    "PC": "PAC a Cobrar",
    "PD": "PAC",
    "PE": "PAC",
    "PF": "Passaporte",
    "PG": "PAC",
    "PH": "PAC",
    "PI": "PAC",
    "PJ": "PAC",
    "PK": "PAC Extra Grande",
    "PL": "PAC",
    "PN": "PAC Normal",
    "PR": "Reembolso Postal",
    "QQ": "Objeto de teste (SIGEP Web)",
    "RA": "Objeto registrado / prioritário",
    "RB": "Carta registrada",
    "RC": "Carta registrada com Valor Declarado",
    "RD": "Remessa econômica ou objeto registrado (DETRAN)",
    "RE": "Objeto registrado econômico",
    "RF": "Receita Federal",
    "RG": "Objeto registrado",
    "RH": "Objeto registrado com AR Digital",
    "RI": "Objeto registrado internacional prioritário",
    "RJ": "Objeto registrado",
    "RK": "Objeto registrado",
    "RL": "Objeto registrado",
    "RM": "Objeto registrado urgente",
    "RN": "Objeto registrado (SIGEPWEB ou Agência)",
    "RO": "Objeto registrado",
    "RP": "Reembolso Postal",
    "RQ": "Objeto registrado",
    "RR": "Objeto registrado",
    "RS": "Objeto registrado",
    "RT": "Remessa econômica Talão/Cartão",
    "RU": "Objeto registrado (ECT)",
    "RV": "Remessa econômica CRLV/CRV/CNH e Notificações com AR Digital",
    "RW": "Objeto internacional",
    "RX": "Objeto internacional",
    "RY": "Remessa econômica Talão/Cartão com AR Digital",
    "RZ": "Objeto registrado",
    "SA": "SEDEX",
    "SB": "SEDEX 10",
    "SC": "SEDEX a cobrar",
    "SD": "SEDEX ou Remessa Expressa (DETRAN)",
    "SE": "SEDEX",
    "SF": "SEDEX",
    "SG": "SEDEX",
    "SH": "SEDEX com AR Digital / SEDEX ou AR Digital",
    "SI": "SEDEX",
    "SJ": "SEDEX Hoje",
    "SK": "SEDEX",
    "SL": "SEDEX",
    "SM": "SEDEX 12",
    "SN": "SEDEX",
    "SO": "SEDEX",
    "SP": "SEDEX Pré-franqueado",
    "SQ": "SEDEX",
    "SR": "SEDEX",
    "SS": "SEDEX",
    "ST": "Remessa Expressa Talão/Cartão",
    "SU": "Encomenda de serviço expressa (ECT)",
    "SV": "Remessa Expressa CRLV/CRV/CNH e Notificações com AR Digital",
    "SW": "e-SEDEX",
    "SX": "SEDEX 10",
    "SY": "Remessa Expressa Talão/Cartão com AR Digital",
    "SZ": "SEDEX",
    "TC": "Objeto para treinamento",
    "TE": "Objeto para treinamento",
    "TS": "Objeto para treinamento",
    "VA": "Encomendas com valor declarado",
    "VC": "Encomendas",
    "VD": "Encomendas com valor declarado",
    "VE": "Encomendas",
    "VF": "Encomendas com valor declarado",
    "VV": "Objeto internacional",
    "XA": "Aviso de chegada (internacional)",
    "XM": "SEDEX Mundi",
    "XR": "Encomenda SUR Postal Expresso",
    "XX": "Encomenda SUR Postal 24 horas",
}

EXTRA_SERVICES = {
    1: {'code': "AR", 'name': "Aviso de Recebimento", 'display_on_label': True},
    2: {'code': "MP", 'name': "Mão Própria Nacional", 'display_on_label': True},
    19: {'code': "VD", 'name': "Valor Declarado (Encomendas)", 'display_on_label': True},  # Sedex
    25: {'code': "RR", 'name': "Registro Nacional", 'display_on_label': False},
    64: {'code': "VD", 'name': "Valor Declarado (Encomendas)", 'display_on_label': True},  # PAC
}

EXTRA_SERVICE_AR = 1
EXTRA_SERVICE_MP = 2
EXTRA_SERVICE_VD_SEDEX = 19
EXTRA_SERVICE_RR = 25
EXTRA_SERVICE_VD_PAC = 64
EXTRA_SERVICE_VD = EXTRA_SERVICE_VD_SEDEX  # For backward compatibility


SERVICES = {
    '40215': {
        'id': 104707,
        'description': 'SEDEX 10',
        'category': 'SERVICO_COM_RESTRICAO',
        'max_weight': 10000,
        'display_name': 'SEDEX 10',
        'symbol': "premium",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '81019': {
        'id': 104672,
        'description': 'E-SEDEX STANDARD',
        'category': 'SERVICO_COM_RESTRICAO',
        'max_weight': 15000,
        'display_name': 'E-SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '41068': {
        'id': 109819,
        'description': 'PAC CONTRATO AGENCIA',
        'category': 'PAC',
        'display_name': 'PAC',
        'max_weight': 30000,
        'symbol': "standard",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("3000.00"),
    },
    '04669': {
        'id': 124884,
        'description': 'PAC',
        'category': 'PAC',
        'display_name': 'PAC',
        'max_weight': 30000,
        'symbol': "standard",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("3000.00"),
    },
    '40444': {
        'id': 109811,
        'description': 'SEDEX - CONTRATO',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40436': {
        'id': 109810,
        'description': 'SEDEX - CONTRATO',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40096': {
        'id': 104625,
        'description': 'SEDEX (CONTRATO)',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '04162': {
        'id': 124849,
        'description': 'SEDEX CONTRATO AGENCIA',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40380': {
        'id': 109806,
        'description': 'SEDEX REVERSO 40096',
        'category': 'REVERSO',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40010': {
        'id': 104295,
        'description': 'SEDEX A VISTA',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '41211': {
        'id': 113546,
        'description': 'PAC - CONTRATO',
        'category': 'PAC',
        'display_name': 'PAC',
        'max_weight': 30000,
        'symbol': "standard",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("3000.00"),
    },
    '40630': {
        'id': 114976,
        'description': 'SEDEX PAGAMENTO NA ENTREGA -',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40916': {
        'id': 118568,
        'description': 'SEDEX AGRUPADO II',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40908': {
        'id': 118567,
        'description': 'SEDEX AGRUPADO I',
        'category': 'SEDEX',
        'max_weight': 30000,
        'display_name': 'SEDEX',
        'symbol': "express",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '41300': {
        'id': 120366,
        'description': 'PAC GRANDES FORMATOS',
        'category': 'SERVICO_COM_RESTRICAO',
        'max_weight': 50000,
        'display_name': 'PAC',
        'symbol': "standard",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("3000.00"),
    },
    '40169': {
        'id': 115218,
        'description': 'SEDEX 12',
        'category': 'SERVICO_COM_RESTRICAO',
        'max_weight': 10000,
        'display_name': 'SEDEX 12',
        'symbol': "premium",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '40290': {
        'id': 108934,
        'description': 'SEDEX HOJE',
        'category': 'SERVICO_COM_RESTRICAO',
        'max_weight': 10000,
        'display_name': 'SEDEX Hoje',
        'symbol': "premium",
        'default_extra_services': [EXTRA_SERVICE_RR],
        'min_declared_value': Decimal("19.50"),
        'max_declared_value': Decimal("10000.00"),
    },
    '10154': {
        'id': 118424,
        'description': 'CARTA COMERCIAL  REGISTRADA',
        'category': 'CARTA_REGISTRADA',
        'display_name': 'Carta Registrada',
    },
    '41246': {
        'id': 115487,
        'description': 'REM. CAMPANHA PAPAI NOEL DOS',
        'category': 'SEM_CATEGORIA',
        'display_name': 'Papai Noel dos Correios'},
    '40150': {
        'id': 115136,
        'description': 'SERVICO DE PROTOCOLO POSTAL -',
        'category': 'SEDEX',
        'display_name': 'Protocolo',
    },
    '10065': {
        'id': 109480,
        'description': 'CARTA COMERCIAL A FATURAR',
        'category': 'CARTA_REGISTRADA',
        'display_name': 'Carta Comercial',
    },
}  # type: Dict[str, dict]

SERVICE_PAC = '04669'
SERVICE_SEDEX = '04162'
SERVICE_SEDEX10 = '40215'
SERVICE_SEDEX12 = '40169'
SERVICE_E_SEDEX = '81019'

INSURANCE_VALUE_THRESHOLD_PAC = Decimal("19.50")  # R$
INSURANCE_VALUE_THRESHOLD_SEDEX = Decimal("19.50")  # R$
INSURANCE_PERCENTUAL_COST = Decimal("0.01")  # 1%

REGIONAL_DIRECTIONS = {
    1: {'code': "AC", 'name': "AC - ADMINISTRAÇAO CENTRAL"},
    3: {'code': "ACR", 'name': "DR - ACRE"},
    4: {'code': "AL", 'name': "DR - ALAGOAS"},
    6: {'code': "AM", 'name': "DR - AMAZONAS"},
    5: {'code': "AP", 'name': "DR - AMAPÁ"},
    8: {'code': "BA", 'name': "DR - BAHIA"},
    10: {'code': "BSB", 'name': "DR - BRASÍLIA"},
    12: {'code': "CE", 'name': "DR - CEARÁ"},
    14: {'code': "ES", 'name': "DR - ESPIRITO SANTO"},
    16: {'code': "GO", 'name': "DR - GOIÁS"},
    18: {'code': "MA", 'name': "DR - MARANHÃO"},
    20: {'code': "MG", 'name': "DR - MINAS GERAIS"},
    22: {'code': "MS", 'name': "DR - MATO GROSSO DO SUL"},
    24: {'code': "MT", 'name': "DR - MATO GROSSO"},
    28: {'code': "PA", 'name': "DR - PARÁ"},
    30: {'code': "PB", 'name': "DR - PARAÍBA"},
    32: {'code': "PE", 'name': "DR - PERNAMBUCO"},
    34: {'code': "PI", 'name': "DR - PIAUÍ"},
    36: {'code': "PR", 'name': "DR - PARANÁ"},
    50: {'code': "RJ", 'name': "DR - RIO DE JANEIRO"},
    60: {'code': "RN", 'name': "DR - RIO GRANDE DO NORTE"},
    26: {'code': "RO", 'name': "DR - RONDONIA"},
    65: {'code': "RR", 'name': "DR - RORAIMA"},
    64: {'code': "RS", 'name': "DR - RIO GRANDE DO SUL"},
    68: {'code': "SC", 'name': "DR - SANTA CATARINA"},
    70: {'code': "SE", 'name': "DR - SERGIPE"},
    74: {'code': "SPI", 'name': "DR - SÃO PAULO INTERIOR"},
    72: {'code': "SPM", 'name': "DR - SÃO PAULO"},
    75: {'code': "TO", 'name': "DR - TOCANTINS"},
}

TRACKING_EVENT_TYPES = {
    "ERROR": "Evento de erro",  # custom event type for "Not Found" error
    "BDE": "Baixa de distribuição externa",
    "BDI": "Baixa de distribuição interna",
    "BDR": "Baixa corretiva",
    "BLQ": "Bloqueio de objetos",
    # "CAR": "Conferência de lista de registro",  # usage not found in SRO documentation
    "CD": "Conferência de nota de despacho",
    "CMT": "Chegada de um meio de transporte",
    "CO": "Coleta de objetos",
    "CUN": "Conferência de lista de registro",
    "DO": "Expedição de nota de despacho",
    "EST": "Estorno",
    "FC": "Função complementar",
    "IDC": "Indenização de objetos",
    "LDI": "Lista de distribuição interna",
    "LDE": "Lista de distribuição externa",
    "OEC": "Lista de Objetos Entregues ao Carteiro",
    "PAR": "Conferência Unidade Internacional",
    "PMT": "Partida Meio de Transporte",
    "PO": "Postagem (exceção)",
    "RO": "Expedição de Lista de Registro",
    "TRI": "Triagem",
    # "CMR": "Evento Desconhecido",  # event type not found in SRO documentation (issued on may, 2017)
}

TRACKING_STATUS = {
    ('ERROR', 0): ('error', 'Objeto não encontrado', '', ''),  # custom status for "Not Found" error
    ('BDE', 0): (
        'delivered',
        'Objeto entregue ao destinatário',
        'Recebido por:',
        'Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.',
    ),
    ('BDI', 0): (
        'delivered',
        'Objeto entregue ao destinatário',
        'Recebido por:',
        'Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.',
    ),
    ('BDR', 0): (
        'delivered',
        'Objeto entregue ao destinatário',
        'Recebido por:',
        'Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.',
    ),
    ('CD', 0): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('CMT', 0): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('CUN', 0): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('DO', 0): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('LDE', 0): (
        'shipped',
        'Objeto saiu para entrega ao remetente',
        '',
        'Acompanhar',
    ),
    ('LDI', 0): (
        'waiting_retrieval',
        'Objeto aguardando retirada no endereço indicado',
        'Endereço:',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('OEC', 0): (
        'shipped',
        'Objeto saiu para entrega ao destinatário',
        '',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('PO', 0): (
        'shipped',
        'Objeto postado',
        '',
        'Acompanhar',
    ),
    ('RO', 0): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('TRI', 0): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('BDE', 1): (
        'delivered',
        'Objeto entregue ao destinatário',
        'Recebido por:',
        'Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.',
    ),
    ('BDI', 1): (
        'delivered',
        'Objeto entregue ao destinatário',
        'Recebido por:',
        'Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.',
    ),
    ('BDR', 1): (
        'delivered',
        'Objeto entregue ao destinatário',
        'Recebido por:',
        'Finalizar a entrega. Não é mais necessário prosseguir com o acompanhamento.',
    ),
    ('BLQ', 1): (
        'shipped',
        'Entrega de objeto bloqueada a pedido do remetente',
        'Objeto em análise de destinação',
        'Acompanhar',
    ),
    ('CD', 1): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('CO', 1): (
        'shipped',
        'Objeto coletado',
        '',
        'Acompanhar',
    ),
    ('CUN', 1): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('DO', 1): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('EST', 1): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('FC', 1): (
        'shipped',
        'Objeto será devolvido por solicitação do remetente',
        '',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('IDC', 1): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('LDI', 1): (
        'waiting_retrieval',
        'Objeto aguardando retirada no endereço indicado',
        'Endereço:',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('OEC', 1): (
        'shipped',
        'Objeto saiu para entrega ao destinatário',
        '',
        'Acompanhar',
    ),
    ('PMT', 1): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('PO', 1): (
        'shipped',
        'Objeto postado',
        '',
        'Acompanhar',
    ),
    ('RO', 1): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('BDE', 2): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Aguarde: Objeto estará disponível para retirada na unidade a ser informada.',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('BDI', 2): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Aguarde: Objeto estará disponível para retirada na unidade a ser informada.',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('BDR', 2): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Aguarde: Objeto estará disponível para retirada na unidade a ser informada.',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('CD', 2): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('DO', 2): (
        'shipped',
        'Objeto encaminhado para',
        '<nome da cidade>',
        'Acompanhar',
    ),
    ('EST', 2): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('FC', 2): (
        'shipped',
        'Objeto com data de entrega agendada',
        '',
        'Acompanhar',
    ),
    ('IDC', 2): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('LDI', 2): (
        'waiting_retrieval',
        'Objeto disponível para retirada em Caixa Postal',
        '',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('BDE', 3): (
        'waiting_retrieval',
        'Remetente não retirou objeto na Unidade dos Correios',
        'Objeto em análise de destinação',
        'Acompanhar. O interessado não buscou o objeto na unidade dos Correios durante o período de guarda.',
    ),
    ('BDI', 3): (
        'waiting_retrieval',
        'Remetente não retirou objeto na Unidade dos Correios',
        'Objeto em análise de destinação',
        'Acompanhar. O interessado não buscou o objeto na unidade dos Correios durante o período de guarda.',
    ),
    ('BDR', 3): (
        'waiting_retrieval',
        'Remetente não retirou objeto na Unidade dos Correios',
        'Objeto em análise de destinação',
        'Acompanhar. O interessado não buscou o objeto na unidade dos Correios durante o período de guarda.',
    ),
    ('CD', 3): (
        'shipped',
        'Objeto recebido na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('EST', 3): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('FC', 3): (
        'shipped',
        'Objeto mal encaminhado',
        'Encaminhamento a ser corrigido.',
        'Acompanhar',
    ),
    ('IDC', 3): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('LDI', 3): (
        'waiting_retrieval',
        'Objeto aguardando retirada no endereço indicado',
        'Endereço:',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('BDE', 4): (
        'delivery_rejected',
        'A entrega não pode ser efetuada - Cliente recusou-se a receber',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 4): (
        'delivery_rejected',
        'A entrega não pode ser efetuada - Cliente recusou-se a receber',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 4): (
        'delivery_rejected',
        'A entrega não pode ser efetuada - Cliente recusou-se a receber',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('EST', 4): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('FC', 4): (
        'shipped',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('IDC', 4): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('BDE', 5): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 5): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 5): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('EST', 5): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('FC', 5): (
        'shipped',
        'Objeto devolvido aos Correios',
        '',
        'Acompanhar',
    ),
    ('IDC', 5): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('BDE', 6): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Cliente desconhecido no local',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 6): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Cliente desconhecido no local',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 6): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Cliente desconhecido no local',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('EST', 6): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('IDC', 6): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('BDE', 7): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('BDI', 7): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('BDR', 7): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('FC', 7): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Empresa sem expediente',
        'A entrega deverá ocorrer no próximo dia útil',
        'Acompanhar',
    ),
    ('IDC', 7): (
        'lost',
        'Objeto não localizado',
        'Houve indenização dos valores correspondentes',
        'Acompanhar',
    ),
    ('BDE', 8): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 8): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 8): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('FC', 8): (
        'shipped',
        'Área com distribuição sujeita a prazo diferenciado',
        'Restrição de entrega domiciliar temporária',
        'Acompanhar',
    ),
    ('BDE', 9): (
        'lost',
        'Objeto não localizado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 9): (
        'lost',
        'Objeto não localizado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 9): (
        'lost',
        'Objeto não localizado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('EST', 9): (
        'ignore_last_entry',
        'Favor desconsiderar a informação anterior',
        '',
        'Acompanhar',
    ),
    ('FC', 9): (
        'delivery_unsuccessful',
        'Remetente não retirou objeto na Unidade dos Correios',
        '',
        'Acompanhar',
    ),
    ('LDE', 9): (
        'shipped',
        'Objeto saiu para entrega ao remetente',
        '',
        'Acompanhar',
    ),
    ('OEC', 9): (
        'shipped',
        'Objeto saiu para entrega ao remetente',
        '',
        'Acompanhar',
    ),
    ('PO', 9): (
        'shipped',
        'Objeto postado após o horário limite da agência',
        'Objeto sujeito a encaminhamento no próximo dia útil',
        'Acompanhar',
    ),
    ('BDE', 10): (
        'shipped_wrong_address',
        'A entrega não pode ser efetuada - Cliente mudou-se',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 10): (
        'shipped_wrong_address',
        'A entrega não pode ser efetuada - Cliente mudou-se',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 10): (
        'shipped_wrong_address',
        'A entrega não pode ser efetuada - Cliente mudou-se',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 12): (
        'waiting_retrieval',
        'Remetente não retirou objeto na Unidade dos Correios',
        'Objeto em análise de destinação',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 12): (
        'waiting_retrieval',
        'Remetente não retirou objeto na Unidade dos Correios',
        'Objeto em análise de destinação',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 12): (
        'waiting_retrieval',
        'Remetente não retirou objeto na Unidade dos Correios',
        'Objeto em análise de destinação',
        'Acionar atendimento dos Correios.',
    ),
    ('LDI', 14): (
        'waiting_retrieval',
        'Objeto aguardando retirada no endereço indicado',
        'Endereço:',
        'Acompanhar. O interessado deverá buscar o objeto em uma Unidade dos Correios.',
    ),
    ('PAR', 15): (
        'delivered',
        'Objeto recebido em <destino>',
        '',
        'Acompanhar',
    ),
    ('PAR', 16): (
        'customs_control',
        'Objeto recebido no Brasil',
        'Objeto sujeito à fiscalização e atraso na entrega',
        'Acompanhar',
    ),
    ('PAR', 17): (
        'customs_control',
        'Objeto liberado pela alfândega',
        '',
        'Acompanhar',
    ),
    ('PAR', 18): (
        'shipped',
        'Objeto recebido na unidade de exportação',
        '',
        'Acompanhar',
    ),
    ('BDE', 19): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 19): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 19): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Endereço incorreto',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 20): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Será realizada nova tentativa de entrega',
        'Acompanhar',
    ),
    ('BDI', 20): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Será realizada nova tentativa de entrega',
        'Acompanhar',
    ),
    ('BDR', 20): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Será realizada nova tentativa de entrega',
        'Acompanhar',
    ),
    ('BDE', 21): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 21): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 21): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Carteiro não atendido',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 22): (
        'shipped',
        'Objeto devolvido aos Correios',
        '',
        'Acompanhar',
    ),
    ('BDI', 22): (
        'shipped',
        'Objeto devolvido aos Correios',
        '',
        'Acompanhar',
    ),
    ('BDR', 22): (
        'shipped',
        'Objeto devolvido aos Correios',
        '',
        'Acompanhar',
    ),
    ('BDE', 23): (
        'returned',
        'Objeto devolvido ao remetente',
        'Recebido por:',
        'Acompanhar',
    ),
    ('BDI', 23): (
        'returned',
        'Objeto devolvido ao remetente',
        'Recebido por:',
        'Acompanhar',
    ),
    ('BDR', 23): (
        'returned',
        'Objeto devolvido ao remetente',
        'Recebido por:',
        'Acompanhar',
    ),
    ('BDE', 24): (
        'waiting_retrieval',
        'Objeto disponível para retirada em Caixa Postal',
        '',
        'Acompanhar',
    ),
    ('BDI', 24): (
        'waiting_retrieval',
        'Objeto disponível para retirada em Caixa Postal',
        '',
        'Acompanhar',
    ),
    ('BDR', 24): (
        'waiting_retrieval',
        'Objeto disponível para retirada em Caixa Postal',
        '',
        'Acompanhar',
    ),
    ('BDE', 25): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Empresa sem expediente',
        'A entrega deverá ocorrer no próximo dia útil',
        'Acompanhar',
    ),
    ('BDI', 25): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Empresa sem expediente',
        'A entrega deverá ocorrer no próximo dia útil',
        'Acompanhar',
    ),
    ('BDR', 25): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Empresa sem expediente',
        'A entrega deverá ocorrer no próximo dia útil',
        'Acompanhar',
    ),
    ('BDE', 26): (
        'waiting_retrieval',
        'Destinatário não retirou objeto na Unidade dos Correios',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 26): (
        'waiting_retrieval',
        'Destinatário não retirou objeto na Unidade dos Correios',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 26): (
        'waiting_retrieval',
        'Destinatário não retirou objeto na Unidade dos Correios',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 28): (
        'damaged',
        'Objeto e/ou conteúdo avariado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 28): (
        'damaged',
        'Objeto e/ou conteúdo avariado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 28): (
        'damaged',
        'Objeto e/ou conteúdo avariado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDE', 32): (
        'shipped',
        'Objeto com data de entrega agendada',
        '',
        'Acompanhar',
    ),
    ('BDI', 32): (
        'shipped',
        'Objeto com data de entrega agendada',
        '',
        'Acompanhar',
    ),
    ('BDR', 32): (
        'shipped',
        'Objeto com data de entrega agendada',
        '',
        'Acompanhar',
    ),
    ('BDE', 33): (
        'delivery_rejected',
        'A entrega não pode ser efetuada - Destinatário não apresentou documento exigido',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 33): (
        'delivery_rejected',
        'A entrega não pode ser efetuada - Destinatário não apresentou documento exigido',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 33): (
        'delivery_rejected',
        'A entrega não pode ser efetuada - Destinatário não apresentou documento exigido',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 34): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Logradouro com numeração irregular',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('BDI', 34): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Logradouro com numeração irregular',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('BDR', 34): (
        'delivery_unsuccessful',
        'A entrega não pode ser efetuada - Logradouro com numeração irregular',
        'Objeto sujeito a atraso na entrega ou a devolução ao remetente',
        'Acompanhar',
    ),
    ('BDE', 35): (
        'delivery_unsuccessful',
        'Coleta ou entrega de objeto não efetuada',
        'Será realizada nova tentativa de coleta ou entrega',
        'Acompanhar',
    ),
    ('BDI', 35): (
        'delivery_unsuccessful',
        'Coleta ou entrega de objeto não efetuada',
        'Será realizada nova tentativa de coleta ou entrega',
        'Acompanhar',
    ),
    ('BDR', 35): (
        'delivery_unsuccessful',
        'Coleta ou entrega de objeto não efetuada',
        'Será realizada nova tentativa de coleta ou entrega',
        'Acompanhar',
    ),
    ('BDE', 36): (
        'delivery_unsuccessful',
        'Coleta ou entrega de objeto não efetuada',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 36): (
        'delivery_unsuccessful',
        'Coleta ou entrega de objeto não efetuada',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 36): (
        'delivery_unsuccessful',
        'Coleta ou entrega de objeto não efetuada',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 37): (
        'damaged',
        'Objeto e/ou conteúdo avariado por acidente com veículo',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 37): (
        'damaged',
        'Objeto e/ou conteúdo avariado por acidente com veículo',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 37): (
        'damaged',
        'Objeto e/ou conteúdo avariado por acidente com veículo',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDE', 38): (
        'delivery_unsuccessful',
        'Objeto endereçado à empresa falida',
        'Objeto será encaminhado para entrega ao administrador judicial',
        'Acompanhar',
    ),
    ('BDI', 38): (
        'delivery_unsuccessful',
        'Objeto endereçado à empresa falida',
        'Objeto será encaminhado para entrega ao administrador judicial',
        'Acompanhar',
    ),
    ('BDR', 38): (
        'delivery_unsuccessful',
        'Objeto endereçado à empresa falida',
        'Objeto será encaminhado para entrega ao administrador judicial',
        'Acompanhar',
    ),
    ('BDE', 40): (
        'customs_control',
        'A importação do objeto/conteúdo não foi autorizada pelos órgãos fiscalizadores',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 40): (
        'customs_control',
        'A importação do objeto/conteúdo não foi autorizada pelos órgãos fiscalizadores',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 40): (
        'customs_control',
        'A importação do objeto/conteúdo não foi autorizada pelos órgãos fiscalizadores',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 41): (
        'unknown',
        'A entrega do objeto está condicionada à composição do lote',
        '',
        'Acompanhar',
    ),
    ('BDI', 41): (
        'unknown',
        'A entrega do objeto está condicionada à composição do lote',
        '',
        'Acompanhar',
    ),
    ('BDR', 41): (
        'unknown',
        'A entrega do objeto está condicionada à composição do lote',
        '',
        'Acompanhar',
    ),
    ('BDE', 42): (
        'unknown',
        'Lote de objetos incompleto',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 42): (
        'unknown',
        'Lote de objetos incompleto',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 42): (
        'unknown',
        'Lote de objetos incompleto',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 43): (
        'customs_control',
        'Objeto apreendido por órgão de fiscalização ou outro órgão anuente',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 43): (
        'customs_control',
        'Objeto apreendido por órgão de fiscalização ou outro órgão anuente',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 43): (
        'customs_control',
        'Objeto apreendido por órgão de fiscalização ou outro órgão anuente',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDE', 45): (
        'shipped',
        'Objeto recebido na unidade de distribuição',
        'Entrega prevista para o próximo dia útil',
        'Acompanhar',
    ),
    ('BDI', 45): (
        'shipped',
        'Objeto recebido na unidade de distribuição',
        'Entrega prevista para o próximo dia útil',
        'Acompanhar',
    ),
    ('BDR', 45): (
        'shipped',
        'Objeto recebido na unidade de distribuição',
        'Entrega prevista para o próximo dia útil',
        'Acompanhar',
    ),
    ('BDE', 46): (
        'delivery_unsuccessful',
        'Tentativa de entrega não efetuada',
        'Entrega prevista para o próximo dia útil',
        'Acompanhar',
    ),
    ('BDI', 46): (
        'delivery_unsuccessful',
        'Tentativa de entrega não efetuada',
        'Entrega prevista para o próximo dia útil',
        'Acompanhar',
    ),
    ('BDR', 46): (
        'delivery_unsuccessful',
        'Tentativa de entrega não efetuada',
        'Entrega prevista para o próximo dia útil',
        'Acompanhar',
    ),
    ('BDE', 47): (
        'delivery_unsuccessful',
        'Saída para entrega cancelada',
        'Será efetuado novo lançamento para entrega',
        'Acompanhar',
    ),
    ('BDI', 47): (
        'delivery_unsuccessful',
        'Saída para entrega cancelada',
        'Será efetuado novo lançamento para entrega',
        'Acompanhar',
    ),
    ('BDR', 47): (
        'delivery_unsuccessful',
        'Saída para entrega cancelada',
        'Será efetuado novo lançamento para entrega',
        'Acompanhar',
    ),
    ('BDE', 48): (
        'delivery_rejected',
        'Retirada em Unidade dos Correios não autorizada pelo remetente',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 48): (
        'delivery_rejected',
        'Retirada em Unidade dos Correios não autorizada pelo remetente',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 48): (
        'delivery_rejected',
        'Retirada em Unidade dos Correios não autorizada pelo remetente',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 49): (
        'delivery_unsuccessful',
        'As dimensões do objeto impossibilitam o tratamento e a entrega',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 49): (
        'delivery_unsuccessful',
        'As dimensões do objeto impossibilitam o tratamento e a entrega',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 49): (
        'delivery_unsuccessful',
        'As dimensões do objeto impossibilitam o tratamento e a entrega',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 50): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 50): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 50): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDE', 51): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 51): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 51): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDE', 52): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDI', 52): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDR', 52): (
        'lost',
        'Objeto roubado',
        'Favor entrar em contato com os Correios.',
        'Acionar atendimento dos Correios.',
    ),
    ('BDE', 53): (
        'shipped',
        'Objeto reimpresso e reenviado',
        '',
        'Acompanhar. O objeto impresso pelos Correios precisou ser refeito e reenviado.',
    ),
    ('BDI', 53): (
        'shipped',
        'Objeto reimpresso e reenviado',
        '',
        'Acompanhar. O objeto impresso pelos Correios precisou ser refeito e reenviado.',
    ),
    ('BDR', 53): (
        'shipped',
        'Objeto reimpresso e reenviado',
        '',
        'Acompanhar. O objeto impresso pelos Correios precisou ser refeito e reenviado.',
    ),
    ('BDE', 54): (
        'customs_control',
        'Para recebimento do objeto, é necessário o pagamento do ICMS Importação',
        '',
        'Acompanhar. O interessado deverá pagar o imposto devido para retirar o objeto em uma Unidade dos Correios.',
    ),
    ('BDI', 54): (
        'customs_control',
        'Para recebimento do objeto, é necessário o pagamento do ICMS Importação',
        '',
        'Acompanhar. O interessado deverá pagar o imposto devido para retirar o objeto em uma Unidade dos Correios.',
    ),
    ('BDR', 54): (
        'customs_control',
        'Para recebimento do objeto, é necessário o pagamento do ICMS Importação',
        '',
        'Acompanhar. O interessado deverá pagar o imposto devido para retirar o objeto em uma Unidade dos Correios.',
    ),
    ('BDE', 55): (
        'customs_control',
        'Solicitada revisão do tributo estabelecido',
        '',
        'Acompanhar',
    ),
    ('BDI', 55): (
        'customs_control',
        'Solicitada revisão do tributo estabelecido',
        '',
        'Acompanhar',
    ),
    ('BDR', 55): (
        'customs_control',
        'Solicitada revisão do tributo estabelecido',
        '',
        'Acompanhar',
    ),
    ('BDE', 56): (
        'customs_control',
        'Declaração aduaneira ausente ou incorreta',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDI', 56): (
        'customs_control',
        'Declaração aduaneira ausente ou incorreta',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDR', 56): (
        'customs_control',
        'Declaração aduaneira ausente ou incorreta',
        'Objeto será devolvido ao remetente',
        'Acompanhar o retorno do objeto ao remetente.',
    ),
    ('BDE', 57): (
        'customs_control',
        'Revisão de tributo concluída - Objeto liberado',
        '',
        'Acompanhar',
    ),
    ('BDI', 57): (
        'customs_control',
        'Revisão de tributo concluída - Objeto liberado',
        '',
        'Acompanhar',
    ),
    ('BDR', 57): (
        'customs_control',
        'Revisão de tributo concluída - Objeto liberado',
        '',
        'Acompanhar',
    ),
    ('BDE', 58): (
        'customs_control',
        'Revisão de tributo concluída - Tributo alterado',
        'O valor do tributo pode ter aumentado ou diminuído',
        'Acompanhar',
    ),
    ('BDI', 58): (
        'customs_control',
        'Revisão de tributo concluída - Tributo alterado',
        'O valor do tributo pode ter aumentado ou diminuído',
        'Acompanhar',
    ),
    ('BDR', 58): (
        'customs_control',
        'Revisão de tributo concluída - Tributo alterado',
        'O valor do tributo pode ter aumentado ou diminuído',
        'Acompanhar',
    ),
    ('BDE', 59): (
        'customs_control',
        'Revisão de tributo concluída - Tributo mantido',
        'Poderá haver incidência de juros e multa.',
        'Acompanhar',
    ),
    ('BDI', 59): (
        'customs_control',
        'Revisão de tributo concluída - Tributo mantido',
        'Poderá haver incidência de juros e multa.',
        'Acompanhar',
    ),
    ('BDR', 59): (
        'customs_control',
        'Revisão de tributo concluída - Tributo mantido',
        'Poderá haver incidência de juros e multa.',
        'Acompanhar',
    ),
    ('BDE', 66): (
        'shipped_delayed',
        'Área com distribuição sujeita a prazo diferenciado',
        'Restrição de entrega domiciliar temporária',
        'Acompanhar',
    ),
    ('BDI', 66): (
        'shipped_delayed',
        'Área com distribuição sujeita a prazo diferenciado',
        'Restrição de entrega domiciliar temporária',
        'Acompanhar',
    ),
    ('BDR', 66): (
        'shipped_delayed',
        'Área com distribuição sujeita a prazo diferenciado',
        'Restrição de entrega domiciliar temporária',
        'Acompanhar',
    ),
    ('BDE', 69): (
        'shipped_delayed',
        'Objeto com atraso na entrega',
        '',
        'Acompanhar',
    ),
    ('BDI', 69): (
        'shipped_delayed',
        'Objeto com atraso na entrega',
        '',
        'Acompanhar',
    ),
    ('BDR', 69): (
        'shipped_delayed',
        'Objeto com atraso na entrega',
        '',
        'Acompanhar',
    ),
    ('BDE', 80): (
        'lost',
        'Objeto extraviado',
        '',
        'Acionar a CAC dos Correios',
    ),
    ('BDI', 80): (
        'lost',
        'Objeto extraviado',
        '',
        'Acionar a CAC dos Correios',
    ),
    ('BDR', 80): (
        'lost',
        'Objeto extraviado',
        '',
        'Acionar a CAC dos Correios',
    ),
}

ZIP_CODE_MAP = {
    'SP': (
        RangeSet((1000, 20000)),  # all
        RangeSet((1000, 6000), (8000, 8500)),  # capital
    ),
    'RJ': (
        RangeSet((20000, 29000)),
        RangeSet((20000, 23800)),
    ),
    'ES': (
        RangeSet((29000, 30000)),
        RangeSet((29000, 29100)),
    ),
    'MG': (
        RangeSet((30000, 40000)),
        RangeSet((30000, 32000)),
    ),
    'BA': (
        RangeSet((40000, 49000)),
        RangeSet((40000, 42600)),
    ),
    'SE': (
        RangeSet((49000, 50000)),
        RangeSet((49000, 49100)),
    ),
    'PE': (
        RangeSet((50000, 57000)),
        RangeSet((50000, 53000)),
    ),
    'AL': (
        RangeSet((57000, 58000)),
        RangeSet((57000, 57100)),
    ),
    'PB': (
        RangeSet((58000, 59000)),
        RangeSet((58000, 58100)),
    ),
    'RN': (
        RangeSet((59000, 60000)),
        RangeSet((59000, 59140)),
    ),
    'CE': (
        RangeSet((60000, 64000)),
        RangeSet((60000, 61600)),
    ),
    'PI': (
        RangeSet((64000, 65000)),
        RangeSet((64000, 64100)),
    ),
    'MA': (
        RangeSet((65000, 66000)),
        RangeSet((65000, 65110)),
    ),
    'PA': (
        RangeSet((66000, 68900)),
        RangeSet((66000, 67000)),
    ),
    'AP': (
        RangeSet((68900, 69000)),
        RangeSet((68900, 68915)),
    ),
    'AM': (
        RangeSet((69000, 69300), (69400, 69900)),
        RangeSet((69000, 69100), (69000, 69100)),
    ),
    'RR': (
        RangeSet((69300, 69400)),
        RangeSet((69300, 69340)),
    ),
    'AC': (
        RangeSet((69900, 70000)),
        RangeSet((69900, 69925)),
    ),
    'DF': (
        RangeSet((70000, 72800), (73000, 73700)),
        RangeSet((70000, 72800), (73000, 73700)),
    ),
    'GO': (
        RangeSet((72800, 73000), (73700, 76800)),
        RangeSet((74000, 74900)),
    ),
    'TO': (
        RangeSet((77000, 78000)),
        RangeSet((77000, 77300)),
    ),
    'MT': (
        RangeSet((78000, 78900)),
        RangeSet((78000, 78110)),
    ),
    'RO': (
        RangeSet((76800, 77000)),
        RangeSet((76800, 76850)),
    ),
    'MS': (
        RangeSet((79000, 80000)),
        RangeSet((79000, 79130)),
    ),
    'PR': (
        RangeSet((80000, 88000)),
        RangeSet((80000, 83000)),
    ),
    'SC': (
        RangeSet((88000, 90000)),
        RangeSet((88000, 88100)),
    ),
    'RS': (
        RangeSet((90000, 100000)),
        RangeSet((90000, 92000)),
    ),
}  # type: Dict[str, Tuple[RangeSet, RangeSet]]

ZIP_CODES = RangeSet(*(p[0] for p in ZIP_CODE_MAP.values()))

FREIGHT_ERROR_RESPONSES = {
    0: 'Processamento com sucesso',
    -1: 'Código de serviço inválido',
    -2: 'CEP de origem inválido',
    -3: 'CEP de destino inválido',
    -4: 'Peso excedido',
    -5: 'O Valor Declarado não deve exceder R$ 10.000,00',
    -6: 'Serviço indisponível para o trecho informado',
    -7: 'O Valor Declarado é obrigatório para este serviço',
    -8: 'Este serviço não aceita Mão Própria',
    -9: 'Este serviço não aceita Aviso de Recebimento',
    -10: 'Precificação indisponível para o trecho informado',
    -11: 'Para definição do preço deverão ser informados, também, o comprimento, '
         'a largura e altura do objeto em centímetros (cm).',
    -12: 'Comprimento inválido.',
    -13: 'Largura inválida.',
    -14: 'Altura inválida.',
    -15: 'O comprimento não pode ser maior que 105 cm.',
    -16: 'A largura não pode ser maior que 105 cm.',
    -17: 'A altura não pode ser maior que 105 cm.',
    -18: 'A altura não pode ser inferior a 2 cm.',
    -20: 'A largura não pode ser inferior a 11 cm.',
    -22: 'O comprimento não pode ser inferior a 16 cm.',
    -23: 'A soma resultante do comprimento + largura + altura não deve superar a 200 cm.',
    -24: 'Comprimento inválido.',
    -25: 'Diâmetro inválido',
    -26: 'Informe o comprimento.',
    -27: 'Informe o diâmetro.',
    -28: 'O comprimento não pode ser maior que 105 cm.',
    -29: 'O diâmetro não pode ser maior que 91 cm.',
    -30: 'O comprimento não pode ser inferior a 18 cm.',
    -31: 'O diâmetro não pode ser inferior a 5 cm.',
    -32: 'A soma resultante do comprimento + o dobro do diâmetro não deve superar a 200 cm.',
    -33: 'Sistema temporariamente fora do ar. Favor tentar mais tarde.',
    -34: 'Código Administrativo ou Senha inválidos.',
    -35: 'Senha incorreta.',
    -36: 'Cliente não possui contrato vigente com os Correios.',
    -37: 'Cliente não possui serviço ativo em seu contrato.',
    -38: 'Serviço indisponível para este código administrativo.',
    -39: 'Peso excedido para o formato envelope',
    -40: ('Para definicao do preco deverao ser informados, tambem, o comprimento e a '
          'largura e altura do objeto em centimetros (cm).'),
    -41: 'O comprimento nao pode ser maior que 60 cm.',
    -42: 'O comprimento nao pode ser inferior a 16 cm.',
    -43: 'A soma resultante do comprimento + largura nao deve superar a 120 cm.',
    -44: 'A largura nao pode ser inferior a 11 cm.',
    -45: 'A largura nao pode ser maior que 60 cm.',
    -888: 'Erro ao calcular a tarifa',
    6: 'Localidade de origem não abrange o serviço informado',
    # 7: 'Localidade de destino não abrange o serviço informado',
    7: 'Serviço indisponível, tente mais tarde',  #
    8: 'Serviço indisponível para o trecho informado',
    9: 'CEP inicial pertencente a Área de Risco.',
    10: 'Área com entrega temporariamente sujeita a prazo diferenciado.',
    11: 'CEP inicial e final pertencentes a Área de Risco',
    99: 'Outros erros diversos do .Net',
}

FREIGHT_ERROR_INITIAL_ZIPCODE_RESTRICTED = 9
FREIGHT_ERROR_FINAL_ZIPCODE_RESTRICTED = 10
FREIGHT_ERROR_INITIAL_AND_FINAL_ZIPCODE_RESTRICTED = 11



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
import warnings
from decimal import Decimal
from typing import List, Tuple, Union

from phonenumbers import NumberParseException, PhoneNumberFormat, format_number, parse

from correios.exceptions import InvalidStateError, InvalidZipCodeError
from correios.models.data import ZIP_CODE_MAP, ZIP_CODES
from correios.utils import capitalize_phrase, rreplace

ZIP_CODE_LENGTH = 8
STATE_LENGTH = 2


class ZipCode:
    REGION_CAPITAL = 'CAPITAL'
    REGION_INTERIOR = 'INTERIOR'
    ALL = 0
    CAPITAL = 1

    def __init__(self, code: str) -> None:
        self._code = self._validate(code)

    @property
    def code(self) -> str:
        return self._code

    def _validate(self, code) -> str:
        if not re.match(r"^\d{5}-?\d{3}$", code):
            raise InvalidZipCodeError("Invalid zipcode {}".format(code))

        code = code.replace("-", "")

        if int(code[:5]) not in ZIP_CODES:
            raise InvalidZipCodeError("Invalid zipcode prefix {}".format(code[:5]))

        return code

    def display(self) -> str:
        return "{}-{}".format(self.code[:5], self.code[-3:])

    def __eq__(self, other):
        if isinstance(other, ZipCode):
            return self.code == other.code

        return self._code == self._validate(other)

    def __str__(self):
        return self.code

    def __repr__(self):
        return "<ZipCode code: {}>".format(self.code)

    def _next_multiple(self, n: int) -> int:
        return n + (n % 10 and (10 - (n % 10)))

    @property
    def digit(self):
        validator = sum(int(d) for d in self.code)
        return self._next_multiple(validator) - validator

    @property
    def prefix(self):
        return int(self.code[:5])

    @property
    def sufix(self):
        return int(self.code[5:])

    @property
    def state(self):
        for state, postal_data in ZIP_CODE_MAP.items():
            if self.prefix in postal_data[self.ALL]:
                return state

    @property
    def region(self):
        postal_data = ZIP_CODE_MAP[self.state]
        if self.prefix in postal_data[self.CAPITAL]:
            return self.REGION_CAPITAL
        return self.REGION_INTERIOR

    @classmethod
    def create(cls, code: Union['ZipCode', int, str]) -> 'ZipCode':
        if isinstance(code, ZipCode):
            return code

        if isinstance(code, int):
            code = str(code)

        return ZipCode(code)


class State:
    STATES = {
        'AC': 'Acre',
        'AL': 'Alagoas',
        'AP': 'Amapá',
        'AM': 'Amazonas',
        'BA': 'Bahia',
        'CE': 'Ceará',
        'DF': 'Distrito Federal',
        'ES': 'Espírito Santo',
        'GO': 'Goiás',
        'MA': 'Maranhão',
        'MT': 'Mato Grosso',
        'MS': 'Mato Grosso do Sul',
        'MG': 'Minas Gerais',
        'PA': 'Pará',
        'PB': 'Paraíba',
        'PR': 'Paraná',
        'PE': 'Pernambuco',
        'PI': 'Piauí',
        'RJ': 'Rio de Janeiro',
        'RN': 'Rio Grande do Norte',
        'RS': 'Rio Grande do Sul',
        'RO': 'Rondônia',
        'RR': 'Roraima',
        'SC': 'Santa Catarina',
        'SP': 'São Paulo',
        'SE': 'Sergipe',
        'TO': 'Tocantins',
    }
    _name_map = {v.lower(): k for k, v in STATES.items()}

    def __init__(self, code: str) -> None:
        self._code = self._validate(code)

    @property
    def code(self) -> str:
        return self._code

    def _validate(self, raw_state) -> str:
        state = self._name_map.get(raw_state.lower(), raw_state)
        state = state.upper()

        if len(state) != STATE_LENGTH or state not in self.STATES:
            raise InvalidStateError("State code {} is invalid".format(state))

        return state

    def display(self):
        return self.STATES[self.code]

    def __eq__(self, other):
        return self.code == self._validate(other)

    def __str__(self):
        return self.code

    def __repr__(self):
        return "<State code: {} name: {}>".format(self.code, self.display())


class ZipAddress:
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 id: int,
                 zip_code: Union[ZipCode, str],
                 state: Union[State, str],
                 city: str,
                 district: str,
                 address: str,
                 complements: List[str]) -> None:
        self.id = id
        self.zip_code = ZipCode(str(zip_code))
        self.state = State(str(state))
        self.city = city
        self.district = district
        self.address = address
        self.complements = [c for c in complements if c]


class Phone:
    def __init__(self, number: str, country="BR") -> None:
        try:
            self.parsed = self._parse(number, country)
        except NumberParseException:
            self.parsed = None
        self.country = country
        self.number = "".join(d for d in number if d.isdigit())

    def _parse(self, number: str, country: str):
        if number.startswith("+"):
            return parse(number)
        return parse(number, country)

    def display(self) -> str:
        if not self.parsed:
            return ""
        country_code = "+{!s}".format(self.parsed.country_code)
        return format_number(self.parsed, PhoneNumberFormat.INTERNATIONAL).replace(country_code, "").strip()

    @property
    def short(self) -> str:
        if not self.parsed:
            return ""
        return str(self.parsed.national_number)

    def __eq__(self, other):
        if not isinstance(other, Phone):
            other = Phone(other, self.country)
        return self.parsed == other.parsed

    def __str__(self):
        if not self.parsed:
            return ""
        return "{}{}".format(self.parsed.country_code, self.parsed.national_number)

    def __repr__(self):
        return "<Phone {!s}>".format(self)


class Address:
    def __init__(self,
                 name: str,
                 street: str,
                 number: Union[int, str],
                 city: str,
                 state: Union[State, str],
                 zip_code: Union[ZipCode, str],
                 complement: str = "",
                 neighborhood: str = "",
                 phone: Union[Phone, str] = "",
                 cellphone: Union[Phone, str] = "",
                 email: str = "",
                 latitude: Union[Decimal, str] = "0.0",
                 longitude: Union[Decimal, str] = "0.0",
                 ) -> None:
        self.name = name
        self.street = street
        self.city = city
        self.complement = complement
        self.neighborhood = neighborhood
        self.email = email
        self.raw_number = str(number)

        if not isinstance(state, State):
            state = State(state)
        self.state = state

        if not isinstance(zip_code, ZipCode):
            zip_code = ZipCode(zip_code)
        self.zip_code = zip_code

        if not isinstance(phone, Phone):
            phone = Phone(phone)
        self.phone = phone

        if not isinstance(cellphone, Phone):
            cellphone = Phone(cellphone)
        self.cellphone = cellphone

        if not isinstance(latitude, Decimal):
            latitude = Decimal(latitude)
        self.latitude = latitude

        if not isinstance(longitude, Decimal):
            longitude = Decimal(longitude)
        self.longitude = longitude

    @property
    def zip_code_display(self) -> str:
        return self.zip_code.display()

    @property
    def basic_address(self) -> str:
        number = self.number
        if self.complement:
            number = "{} - {}".format(self.number, self.complement)

        if self.neighborhood:
            return capitalize_phrase("{}, {}, {}".format(self.street, number, self.neighborhood))
        return capitalize_phrase("{}, {}".format(self.street, number))

    @property
    def label_address(self) -> str:
        msg = "{}.label_address is going to be deprecated. Make sure you use SendAddress or ReceiverAddress"
        warnings.warn(msg.format(type(self).__name__), DeprecationWarning)

        template = ("{address.street!s:>.40} {address.number!s:>.8}<br/> "
                    "{address.complement!s:>.20} {address.neighborhood!s:>.28}")
        return capitalize_phrase(template.format(address=self))

    @property
    def label_name(self) -> str:
        return capitalize_phrase(self.name)

    @property
    def display_address(self) -> Tuple[str, str]:
        address = "{}, {} - {}".format(self.street, self.raw_number, self.complement)
        city = "{} / {} - {}".format(self.city, self.state, self.zip_code.display())
        return address.strip(), city.strip()

    @property
    def filtered_number(self) -> str:
        return "".join(d for d in self.raw_number if d.isdigit())

    @property
    def number(self) -> str:
        return self.filtered_number or "S/N"

    @property
    def zip_complement(self) -> str:
        return self.filtered_number or "0"


class ReceiverAddress(Address):
    @property
    def label_address(self) -> str:
        label_address = self.basic_address

        if len(label_address) <= 55:
            label_address = rreplace(label_address, ',', '<br/>', count=1)

        return label_address


class SenderAddress(Address):
    @property
    def label_address(self) -> str:
        label_address = self.basic_address

        if len(label_address) <= 60:
            label_address = rreplace(label_address, ',', '<br/>', count=1)

        return label_address



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from io import BytesIO
from typing import List  # noqa: F401

from reportlab.graphics.barcode import createBarcodeDrawing
from reportlab.lib import colors, pagesizes
from reportlab.lib.enums import TA_CENTER
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import mm
from reportlab.lib.utils import ImageReader
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.pdfgen.canvas import Canvas
from reportlab.platypus import Flowable, Paragraph, Table, TableStyle

from correios.exceptions import RendererError
from correios.models.data import EXTRA_SERVICE_AR, EXTRA_SERVICE_MP
from correios.models.posting import PostingList, ShippingLabel
from correios.models.user import ExtraService

VERTICAL_SECURITY_MARGIN = 6  # pt


class PDF:
    def __init__(self, page_size):
        self._file = BytesIO()
        self.canvas = Canvas(self._file, pagesize=page_size)
        self._saved = False

    @property
    def file(self) -> BytesIO:
        if not self._saved:
            self.canvas.save()
            self._saved = True

        self._file.seek(0)
        return self._file

    def save(self, filename):
        with open(filename, "wb") as pdf_file:
            pdf_file.write(self.file.read())

    def __bytes__(self):
        return self.file.getvalue()


class ShippingLabelFlowable(Flowable):
    label_style = ParagraphStyle(name="label", fontName="Helvetica", fontSize=6, leading=8)

    def __init__(self,
                 shipping_label: ShippingLabel,
                 label_width,
                 label_height,
                 hmargin=5 * mm,
                 vmargin=5 * mm) -> None:
        super().__init__()
        self.shipping_label = shipping_label

        self.label_width = label_width
        self.hmargin = hmargin
        self.width = self.label_width - (2 * hmargin)

        self.label_height = label_height
        self.vmargin = vmargin
        self.height = self.label_height - (2 * vmargin)

        self.x1 = self.hmargin
        self.y1 = self.vmargin
        self.x2 = self.hmargin + self.width
        self.y2 = self.vmargin + self.height

    def wrap(self, *args):
        return self.label_width, self.label_height

    def draw(self):
        canvas = self.canv
        canvas.setLineWidth(0.1)

        # logo
        logo = ImageReader(self.shipping_label.logo)
        canvas.drawImage(logo, self.x1 + 5 * mm, self.y2 - (27 * mm), width=25 * mm,
                         preserveAspectRatio=True, anchor="sw", mask="auto")

        # datagrid
        datagrid = createBarcodeDrawing("ECC200DataMatrix",
                                        value=self.shipping_label.get_datamatrix_info(),
                                        width=25 * mm, height=25 * mm)
        datagrid.drawOn(canvas, self.x1 + 40 * mm, self.y2 - (27 * mm))

        # symbol
        symbol = ImageReader(self.shipping_label.symbol)
        canvas.drawImage(symbol, self.x1 + 70 * mm, self.y2 - (27 * mm), width=25 * mm,
                         preserveAspectRatio=True, anchor="sw", mask="auto")

        # header shipping_labels
        text = Paragraph("{}<br/>{}".format(self.shipping_label.get_invoice(),
                                            self.shipping_label.get_order()),
                         style=self.label_style)
        text.wrap(25 * mm, 14)
        text.drawOn(canvas, self.x1 + 5 * mm, self.y2 - (28 * mm) - 14)

        text = Paragraph("{}<br/>{}".format(self.shipping_label.get_contract_number(),
                                            self.shipping_label.get_service_name()),
                         style=self.label_style)
        text.wrap(25 * mm, 14)
        text.drawOn(canvas, self.x1 + 40 * mm, self.y2 - (28 * mm) - 14)

        text = Paragraph("{}<br/>{}".format(self.shipping_label.get_package_sequence(),
                                            self.shipping_label.get_weight()),
                         style=self.label_style)
        text.wrap(25 * mm, 14)
        text.drawOn(canvas, self.x1 + 70 * mm, self.y2 - (28 * mm) - 14)

        # tracking
        canvas.setFont("Helvetica-Bold", 10)
        canvas.drawCentredString(self.x1 + 42.5 * mm, self.y2 - 40 * mm,
                                 self.shipping_label.get_tracking_code())
        code = createBarcodeDrawing("Code128", value=str(self.shipping_label.tracking_code),
                                    width=75 * mm, height=18 * mm, quiet=0)
        code.drawOn(canvas, 10 * mm, self.y2 - (59 * mm))

        # extra services (first three)
        first_row = self.y2 - (40 * mm) - 10  # font-size=10pt
        for extra_service in self.shipping_label.extra_services:
            if not extra_service.display_on_label:
                continue

            canvas.drawString(self.x2 - (10 * mm), first_row, extra_service.code)
            first_row -= 14

        # receipt
        receipt_style = ParagraphStyle(name="label", fontName="Helvetica", fontSize=8, leading=14)
        text = Paragraph(self.shipping_label.receipt_template, style=receipt_style)
        text.wrap(self.width - (10 * mm), 28)
        text.drawOn(canvas, self.x1 + (5 * mm), self.y2 - (60 * mm) - 28)

        # sender header
        canvas.setFillColor(colors.black)
        canvas.line(self.x1, self.y2 - (70 * mm), self.x2, self.y2 - (70 * mm))
        width = stringWidth(self.shipping_label.sender_header, "Helvetica", 10) + 2
        canvas.rect(self.x1, self.y2 - (70 * mm) - 14, width, 14, fill=True)

        canvas.setFont("Helvetica", 9)
        canvas.setFillColor(colors.white)
        canvas.drawString(self.x1 + 4, self.y2 - (70 * mm) - 10, self.shipping_label.sender_header)

        carrier_logo = ImageReader(self.shipping_label.carrier_logo)
        canvas.drawImage(carrier_logo, self.x2 - 20 * mm, self.y2 - (70 * mm) - 12, height=10,
                         preserveAspectRatio=True, anchor="sw", mask="auto")

        # receiver
        receiver_style = ParagraphStyle("receiver", fontName="Helvetica", fontSize=10, leading=15)
        text = Paragraph(self.shipping_label.get_receiver_data(), style=receiver_style)
        text.wrap(self.width - 5 * mm, 24 * mm)
        text.drawOn(canvas, self.x1 + 5 * mm, self.y2 - (98 * mm))

        # receiver zip barcode
        code = createBarcodeDrawing("Code128", value=str(self.shipping_label.receiver.zip_code),
                                    width=30 * mm, height=18 * mm, quiet=0)
        code.drawOn(canvas, self.x1 + 5 * mm, self.y2 - (117 * mm))

        # text
        text_style = ParagraphStyle("text", fontName="Helvetica", fontSize=10)
        text = Paragraph(self.shipping_label.text, style=text_style)
        width = self.x2 - (self.x1 + (45 * mm))
        text.wrap(width, 15 * mm)
        text.breakLines(width)
        text.drawOn(canvas, self.x1 + (45 * mm), self.y2 - (98 * mm))

        # sender
        canvas.line(self.x1, self.y2 - (118 * mm), self.x2, self.y2 - (118 * mm))
        sender_style = ParagraphStyle("sender", fontName="Helvetica", fontSize=9)
        text = Paragraph(self.shipping_label.get_sender_data(), style=sender_style)
        text.wrap(self.width - 5 * mm, 22 * mm)
        text.drawOn(canvas, self.x1 + 5 * mm, self.y1 + 2 * mm)

        # border
        canvas.rect(self.x1, self.y1, self.width, self.height)


class PostingReportPDFRenderer:
    label_style = ParagraphStyle(name="label", fontName="Helvetica", fontSize=8, leading=5 * mm)
    table_header_style = ParagraphStyle(name="th", fontName="Courier-Bold", fontSize=8, alignment=TA_CENTER)
    signature_style = ParagraphStyle(name="sign", fontName="Helvetica", fontSize=8, leading=10, alignment=TA_CENTER)

    heading_title = "Lista de Postagem".upper()
    header_label_col1 = ("<b>N° da Lista:</b> {!s}<br/>"
                         "<b>Contrato:</b> {!s}<br/>"
                         "<b>Cód. Administrativo:</b> {!s}<br/>"
                         "<b>Cartão:</b> {!s}")
    header_label_col2 = ("<b>Remetente:</b> {!s}<br/>"
                         "<b>Cliente:</b> {!s}<br/>"
                         "<b>Endereço:</b> {!s}<br/>"
                         "{!s}")
    header_label_col3 = "<b>Telefone:</b> {!s}<br/>"

    table_style = [
        ('FONTNAME', (0, 0), (-1, -1), "Courier"),
        ('FONTNAME', (0, 0), (-1, 0), "Courier-Bold"),
        ('FONTSIZE', (0, 0), (-1, -1), 8),
        ('ALIGN', (6, 1), (6, -1), "RIGHT"),
    ]
    col_widths = (25 * mm, 15 * mm, 10 * mm, 6 * mm, 6 * mm, 6 * mm, 20 * mm, 15 * mm, 13 * mm, 84 * mm)
    max_receiver_name_size = 48  # chars
    table_header = ("Nº do Objeto", "CEP", "Peso", "AR", "MP", "VD",
                    Paragraph("Valor<br/>Declarado", style=table_header_style),
                    Paragraph("Nota<br/>Fiscal", style=table_header_style),
                    "Volume", "Destinatário")
    table_max_rows = 29
    yes, no = "S", "N"

    footer_title_text = "Apresentar esta lista em caso de pedido de informações".upper()
    footer_disclaimer = "Estou ciente do disposto na clásula terceria do contrato de prestação de serviços"
    footer_stamp_text = "Carimbo e assinatura / Matrícula dos Correios"
    footer_signature_text = ("_________________________________________________________<br/>"
                             "ASSINATURA REMETENTE<br/><br/><br/>"
                             "Obs: 1o via p/ a Unidade de Postagem e 2o via p/ o cliente")

    def __init__(self, page_size=pagesizes.A4, shipping_labels_margin=(0, 0), posting_list_margin=(5 * mm, 5 * mm)):
        self.shipping_labels = []  # type: List[ShippingLabel]
        self._tracking_codes = set()

        self.page_size = page_size
        self.page_width = page_size[0]
        self.page_height = page_size[1]

        self.posting_list = None  # type: PostingList
        self.posting_list_margin = posting_list_margin

        self.shipping_labels_margin = shipping_labels_margin
        self.shipping_labels_width = self.page_width - (2 * shipping_labels_margin[0])
        self.shipping_labels_height = self.page_height - (2 * shipping_labels_margin[1])

        self.col_size = self.shipping_labels_width / 2
        self.row_size = self.shipping_labels_height / 2
        self._label_position = (
            (shipping_labels_margin[0], self.page_height / 2),
            (shipping_labels_margin[0] + self.col_size, self.page_height / 2),
            (shipping_labels_margin[0], shipping_labels_margin[1]),
            (shipping_labels_margin[0] + self.col_size, shipping_labels_margin[1]),
        )

    def set_posting_list(self, posting_list: PostingList):
        if not posting_list.closed:
            raise RendererError("Cannot render an open posting list")

        self.posting_list = posting_list
        for shipping_label in posting_list.shipping_labels.values():
            self.add_shipping_label(shipping_label)

    def add_shipping_label(self, shipping_label: ShippingLabel):
        if shipping_label in self.shipping_labels:
            raise RendererError("Shipping Label {!s} already added".format(shipping_label.tracking_code))
        self.shipping_labels.append(shipping_label)

    def draw_grid(self, canvas):
        canvas.setLineWidth(0.2)
        canvas.setStrokeColor(colors.gray)
        canvas.line(self.shipping_labels_margin[0],
                    self.page_height / 2,
                    self.shipping_labels_width + self.shipping_labels_margin[0],
                    self.page_height / 2)
        canvas.line(self.page_width / 2,
                    self.shipping_labels_margin[1],
                    self.page_width / 2,
                    self.shipping_labels_height + self.shipping_labels_margin[1])

    # noinspection PyUnusedLocal
    def _posting_list_header(self, pdf, width, x1, y1, x2, y2):
        canvas = pdf.canvas

        # logo
        logo = ImageReader(self.posting_list.logo)
        canvas.drawImage(logo, x1, y2 - 10.3 * mm, height=8 * mm,
                         preserveAspectRatio=True, anchor="sw", mask="auto")
        # head1
        canvas.setFont("Helvetica-Bold", size=14)
        canvas.drawCentredString(x2 - ((width - 40 * mm) / 2), y2 - 9 * mm,
                                 "Empresa Brasileira de Correios e Telégrafos".upper())
        # box
        canvas.setLineWidth(0.5)
        canvas.rect(x1, y2 - 45 * mm, width, 30 * mm)
        canvas.drawCentredString(x1 + width / 2, y2 - (15 * mm) - 15, self.heading_title)
        # header info
        spacer = 5 * mm
        col_width = width / 4
        col = 0
        header = self.header_label_col1.format(self.posting_list.number,
                                               self.posting_list.contract,
                                               self.posting_list.posting_card.administrative_code,
                                               self.posting_list.posting_card)
        text = Paragraph(header, style=self.label_style)
        text.wrap(col_width, 30 * mm)
        text.drawOn(canvas, x1 + spacer + col * col_width, y2 - (43 * mm))
        col = 1
        header = self.header_label_col2.format(self.posting_list.sender.name[:30],
                                               self.posting_list.contract.customer_name[:30],
                                               self.posting_list.sender.display_address[0][:30],
                                               self.posting_list.sender.display_address[1][:30])
        text = Paragraph(header, style=self.label_style)
        text.wrap(col_width * 2, 30 * mm)
        text.drawOn(canvas, x1 + spacer + col * col_width, y2 - (43 * mm))
        col = 3
        header = self.header_label_col3.format(self.posting_list.sender.phone.display())
        text = Paragraph(header, style=self.label_style)
        text.wrap(col_width, 30 * mm)
        text.drawOn(canvas, x1 + spacer + col * col_width, y2 - (43 * mm))
        code = createBarcodeDrawing("Code128", value=str(self.posting_list.number),
                                    width=col_width * 0.6, height=10 * mm, quiet=0)
        code.drawOn(canvas, x1 + spacer + col * col_width, y2 - (35 * mm))

    # noinspection PyUnusedLocal
    def _posting_list_footer(self, pdf, width, x1, y1, x2, y2):
        canvas = pdf.canvas

        canvas.rect(x1, y1, width, 38 * mm)
        canvas.setFont("Helvetica-Bold", size=9)
        canvas.drawCentredString(x2 - (width / 2), y1 + 38 * mm - 10, self.footer_title_text)
        canvas.setFont("Helvetica", size=8)
        canvas.drawString(x1 + 2 * mm, y1 + 28 * mm, self.footer_disclaimer)
        text_width = stringWidth(self.footer_stamp_text, "Helvetica", 8)
        canvas.drawString(x2 - 2 * mm - text_width, y1 + 28 * mm, self.footer_stamp_text)
        text = Paragraph(self.footer_signature_text, style=self.signature_style)
        text.wrap(stringWidth(self.footer_disclaimer, "Helvetica", 8), 10 * mm)
        text.drawOn(canvas, x1 + 2 * mm, y1 + 2 * mm)

    # noinspection PyUnusedLocal
    def _posting_list_table(self, canvas, x1, y1, x2, y2, shipping_labels):
        style = self.table_style[:]
        table = [self.table_header]
        for i, shipping_label in enumerate(shipping_labels, start=1):
            row = (
                str(shipping_label.tracking_code),
                str(shipping_label.receiver.zip_code),
                str(shipping_label.package.posting_weight),
                self.yes if ExtraService.get(EXTRA_SERVICE_AR) in shipping_label else self.no,
                self.yes if ExtraService.get(EXTRA_SERVICE_MP) in shipping_label else self.no,
                self.yes if shipping_label.has_declared_value() else self.no,
                str(shipping_label.value).replace(".", ",") if shipping_label.value is not None else "",
                str(shipping_label.invoice_number),
                shipping_label.get_package_sequence(),
                shipping_label.receiver.name[:self.max_receiver_name_size],
            )

            # noinspection PyTypeChecker
            table.append(row)

            if i % 2:
                style.append(('BACKGROUND', (0, i), (-1, i), colors.lightgrey))

        table_flow = Table(
            table,
            colWidths=self.col_widths,
            style=TableStyle(style),
        )
        w, h = table_flow.wrap(0, 0)
        table_flow.drawOn(canvas, x1, y2 - h - 50 * mm)

    def render_posting_list(self, pdf=None) -> PDF:
        if pdf is None:
            pdf = PDF(self.page_size)
        canvas = pdf.canvas

        width = self.page_width - (2 * self.posting_list_margin[0])
        height = self.page_height - (2 * self.posting_list_margin[1])
        x1, y1 = self.posting_list_margin[0], self.posting_list_margin[1]
        x2, y2 = width + self.posting_list_margin[0], height + self.posting_list_margin[1]

        for start_group_pos in range(0, len(self.shipping_labels), self.table_max_rows):
            self._posting_list_header(pdf, width, x1, y1, x2, y2)

            group = self.shipping_labels[start_group_pos:start_group_pos + self.table_max_rows]
            self._posting_list_table(canvas, x1, y1, x2, y2, group)

            self._posting_list_footer(pdf, width, x1, y1, x2, y2)
            canvas.showPage()

        return pdf

    def render_labels(self, pdf=None) -> PDF:
        if pdf is None:
            pdf = PDF(self.page_size)

        canvas = pdf.canvas

        position = len(self._label_position) - 1
        for i, shipping_label in enumerate(self.shipping_labels):
            position = i % len(self._label_position)
            self.render_label(shipping_label, position, pdf)
            if position == len(self._label_position) - 1:
                self.draw_grid(canvas)
                canvas.showPage()

        if position != len(self._label_position) - 1:
            self.draw_grid(canvas)
            canvas.showPage()

        return pdf

    def render_label(self, shipping_label: ShippingLabel, position=0, pdf=None) -> PDF:
        single_label = pdf is None
        if single_label:
            pdf = PDF(self.page_size)

        canvas = pdf.canvas
        label = ShippingLabelFlowable(shipping_label, self.col_size, self.row_size)
        label.drawOn(pdf.canvas, *self._label_position[position])

        if single_label:
            self.draw_grid(canvas)
            canvas.showPage()

        return pdf

    def render(self) -> PDF:
        pdf = PDF(self.page_size)
        self.render_posting_list(pdf)
        self.render_labels(pdf)
        return pdf



import pytest

try:
    from correios import xml_utils
except ImportError:
    xml_utils = None


@pytest.mark.skipif(not xml_utils, reason="API Client support disabled")
def test_cdata_special_chars_handling():
    document = xml_utils.Element("test", cdata="áéíóú")
    xml = xml_utils.tostring(document, encoding="unicode").encode("iso-8859-1")
    assert xml == "<test><![CDATA[áéíóú]]></test>".encode("iso-8859-1")



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re

from vcr import VCR

USER_REGEX = re.compile(r'<usuario>\w+</usuario>')
PASS_REGEX = re.compile(r'<senha>.*</senha>')


def replace_auth(request):
    if not request.body:
        return request

    body = request.body.decode()
    body = USER_REGEX.sub(r'<usuario>teste</usuario>', body)
    body = PASS_REGEX.sub(r'<senha>****</senha>', body)
    request.body = body.encode()
    return request


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures")

vcr = VCR(
    record_mode='once',
    serializer='yaml',
    cassette_library_dir=os.path.join(FIXTURES_DIR, 'cassettes'),
    path_transformer=VCR.ensure_suffix('.yaml'),
    match_on=['method'],
    before_record_request=replace_auth,
)



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import find_packages, setup


def load_requirements(filename):
    install_requires = {}
    section = "install_requires"

    reqs = open(filename)
    for raw_line in reqs:
        line = raw_line.strip()
        if not line:
            continue

        if line.startswith("#"):
            section = line.replace("#", "").strip()
            continue

        install_requires.setdefault(section, []).append(line)

    return install_requires


requirements = load_requirements("requirements.txt")

setup(
    name="correios",
    version="6.1.1",
    url="https://github.com/osantana/correios",

    author="Osvaldo Santana Neto",
    author_email="correiospy@osantana.me",

    description="A client library for Brazilian Correios APIs and services",
    long_description=open("README.rst").read(),

    packages=find_packages(),
    include_package_data=True,

    install_requires=requirements.pop("install_requires"),
    extras_require=requirements,

    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],

    entry_points={
        "console_scripts": ["update-correios-wsdl=correios.update_wsdl:cli"],
    },
)



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import tempfile
from unittest import mock
from unittest.mock import call

from requests import HTTPError

from correios.update_wsdl import WSDLUpdater

from .vcr import vcr


def get_wsdl_updater(path):
    response_mock = mock.MagicMock()
    response_mock.text = "response value"
    session = mock.MagicMock()
    session.get.return_value = response_mock

    return WSDLUpdater(wsdl_path=path, session=session)


@vcr.use_cassette
def test_basic_update():
    with tempfile.TemporaryDirectory() as wsdl_path:
        wsdl_updater = WSDLUpdater(wsdl_path=wsdl_path)
        wsdl_updater.update_all()


def test_wsdl_updater_download():
    with tempfile.TemporaryDirectory() as wsdl_path:
        wsdl_updater = get_wsdl_updater(wsdl_path)
        wsdl_updater.download_wsdl('https://example.com/service.wsdl', 'service-production.wsdl')

    wsdl_updater.session.get.assert_called_once_with('https://example.com/service.wsdl')

    printed = wsdl_updater.output.getvalue()
    assert 'Updating file: service-production.wsdl URL: https://example.com/service.wsdl' in printed
    assert 'Successfully create file:' in printed
    assert 'Updated with success' in printed


def test_wsdl_update_all():
    with tempfile.TemporaryDirectory() as wsdl_path:
        wsdl_updater = get_wsdl_updater(wsdl_path)
        wsdl_updater.update_all()
        wsdl_updater.session.get.assert_has_calls([
            call('https://apps.correios.com.br/SigepMasterJPA/AtendeClienteService/AtendeCliente?wsdl'),
            call('https://apphom.correios.com.br/SigepMasterJPA/AtendeClienteService/AtendeCliente?wsdl'),
            call('https://webservice.correios.com.br/service/rastro/Rastro.wsdl'),
            call('http://ws.correios.com.br/calculador/CalcPrecoPrazo.asmx?WSDL'),
        ], any_order=True)


def test_wsdl_download_error():
    response_mock = mock.MagicMock()
    response_mock.text = "response value"
    session = mock.MagicMock()
    session.get.side_effect = HTTPError("test error")

    wsdl_updater = WSDLUpdater(wsdl_path="", session=session)
    wsdl_updater.update_all()

    printed = wsdl_updater.output.getvalue()
    assert ('Updating file: AtendeCliente-production.wsdl '
            'URL: https://apps.correios.com.br/SigepMasterJPA/AtendeClienteService/AtendeCliente?wsdl') in printed
    assert ("Error downloading https://apps.correios.com.br/SigepMasterJPA/AtendeClienteService/AtendeCliente?wsdl: "
            "test error") in printed



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BaseCorreiosError(Exception):
    pass


class ModelError(BaseCorreiosError):
    pass


class ClientError(BaseCorreiosError):
    pass


class InvalidZipCodeError(ModelError):
    pass


class InvalidStateError(ModelError):
    pass


class InvalidFederalTaxNumberError(ModelError):
    pass


class InvalidStateTaxNumberError(ModelError):
    pass


class InvalidUserContractError(ModelError):
    pass


class InvalidTrackingCodeError(ModelError):
    pass


class InvalidShippingLabelError(ModelError):
    pass


class InvalidPackageError(ModelError):
    pass


class InvalidPackageDimensionsError(InvalidPackageError):
    pass


class InvalidMinPackageDimensionsError(InvalidPackageDimensionsError):
    pass


class InvalidMaxPackageDimensionsError(InvalidPackageDimensionsError):
    pass


class InvalidPackageWeightError(InvalidPackageError):
    pass


class InvalidMinPackageWeightError(InvalidPackageWeightError):
    pass


class InvalidMaxPackageWeightError(InvalidPackageWeightError):
    pass


class InvalidPackageSequenceError(InvalidPackageError):
    pass


class InvalidAddressesError(ModelError):
    pass


class InvalidExtraServiceError(ModelError):
    pass


class PostingListError(ModelError):
    pass


class InvalidRegionalDirectionError(ModelError):
    pass


class PostingListSerializerError(ClientError):
    pass


class TrackingCodesLimitExceededError(ClientError):
    pass


class AuthenticationError(ClientError):
    pass


class ConnectTimeoutError(ClientError):
    pass


class ClosePostingListError(ClientError):
    pass


class CanceledPostingCardError(ClientError):
    pass


class NonexistentPostingCardError(ClientError):
    pass


class RendererError(BaseCorreiosError):
    pass


class TrackingCodeNotFoundError(ModelError):
    pass


class InvalidEventStatusError(ModelError):
    pass


class InvalidDeclaredValueError(ModelError):
    pass


class MaximumDeclaredValueError(InvalidDeclaredValueError):
    pass


class MinimumDeclaredValueError(InvalidDeclaredValueError):
    pass
